import contextlib
import os

from PyQt6.QtCore import QThread, pyqtSignal

from parsagon.exceptions import APIException
from parsagon.settings import get_save_api_key_interactive


class GUIController(QThread):
    """
    The GUI controller is a background thread that runs the CLI if the GUI is enabled.  It then delegates all interaction between the CLI code and the GUI.  It is designed as a singleton so that references to the controller don't have to be passed around.
    """

    _instance = None

    update_text_signal = pyqtSignal(str, str, str)  # Include text color
    request_input_signal = pyqtSignal()
    set_loading_signal = pyqtSignal(bool)
    show_progress_signal = pyqtSignal(bool)
    set_progress_title_signal = pyqtSignal(str)
    set_progress_signal = pyqtSignal(int)

    @classmethod
    def shared(cls):
        if cls._instance is None:
            raise RuntimeError("GUIController has not been initialized")
        return cls._instance

    def __init__(self, condition, parent=None):
        super().__init__(parent)
        self.timer = None
        self.condition = condition
        self.current_input = None
        self.progress_total = None
        assert self.__class__._instance is None, "GUIController is a singleton"
        self.__class__._instance = self

    def run(self):
        self.__class__._instance._instance = self
        while True:
            retry = False
            try:
                from parsagon.assistant import assist
                from parsagon.settings import get_graphic, get_api_key

                _ = get_api_key(interactive=True)
                assist(True)
            except Exception as e:
                if isinstance(e, (KeyboardInterrupt, SystemExit)):
                    raise
                else:
                    from parsagon.print import error_print

                    error_print(str(e))
                    error_print(
                        "Parsagon encountered an error.  Please restart the application to continue.  You may want to copy any messages above you want to save."
                    )

                    if isinstance(e, APIException) and e.status_code == 401:
                        get_save_api_key_interactive()
                        retry = True
            if not retry:
                break

    def input(self, prompt):
        if prompt:
            self.print(prompt)
        self.request_input_signal.emit()
        with self.condition:
            self.condition.wait()
            return self.current_input

    def print(self, text, color="black", background="#CBCED2"):
        html = text.replace("\n", "<br>")
        self.update_text_signal.emit(html, color, background)

    @contextlib.contextmanager
    def spinner(self):
        self.set_loading_signal.emit(True)  # Show spinner
        try:
            yield
        finally:
            self.set_loading_signal.emit(False)  # Hide spinner

    def show_progress(self, show_progress, text=""):
        self.show_progress_signal.emit(show_progress)
        if show_progress:
            self.set_progress_title_signal.emit(text)

    def set_progress(self, progress_iteration_num):
        self.set_progress_signal.emit(int(progress_iteration_num / self.progress_total * 100))



import json
from parsagon.api import send_assistant_message, send_assistant_function_outputs, schedule, delete_schedule
from parsagon.create import create_program
from parsagon.executor import Executor
from parsagon.print import assistant_print, assistant_spinner, browser_print, error_print, ask, input, ask_reply
from parsagon.runs import run, batch_runs


def assist(verbose=False):
    task = ask("Type what do you want to do")
    with assistant_spinner():
        response = send_assistant_message(task)
    while True:
        if not response["success"]:
            error_print(
                "The OpenAI API is currently experiencing difficulties. You can try again later, or you can run `parsagon create --no_assistant` to use Parsagon without assistance from GPT."
            )
            return
        for message in response["messages"]:
            if message["role"] != "assistant":
                continue
            for content in message["content"]:
                assistant_print(content["text"]["value"])
        if response["status"] == "completed":
            reply = ask_reply()
            if reply.strip() == "Q":
                break
            with assistant_spinner():
                response = send_assistant_message(reply, response["thread_id"])
        elif response["status"] == "requires_action":
            outputs = []
            for function_call in response["required_action"]:
                name = function_call["function"]["name"]
                args = json.loads(function_call["function"]["arguments"])
                output = {"tool_call_id": function_call["id"], "function_name": name}
                if name == "investigate_page":
                    html = get_page_html(**args)
                    output["output"] = html
                    outputs.append(output)
                elif name == "create_program":
                    result = create_program(args["description"])
                    output["output"] = json.dumps(result)
                    outputs.append(output)
                elif name == "run_program":
                    result = run(**args)
                    output["output"] = json.dumps(result)
                    outputs.append(output)
                elif name == "batch_runs":
                    batch_name = input("Please enter a name for the batch run (for saving of intermediate results): ")
                    result = batch_runs(batch_name, **args)
                    output["output"] = json.dumps(result)
                    outputs.append(output)
                elif name == "set_schedule":
                    result = schedule(**args)
                    output["output"] = json.dumps(result)
                    outputs.append(output)
                elif name == "clear_schedule":
                    delete_schedule(**args)
                    output["output"] = "Schedule cleared"
                    outputs.append(output)
            with assistant_spinner():
                response = send_assistant_function_outputs(outputs, response["thread_id"], response["run_id"])
        else:
            error_print("An error occurred")
            break


def get_page_html(url, headless=False, use_uc=False, page_load_timeout=None, script_timeout=None):
    browser_print(f"Checking what {url} looks like...")
    executor = Executor(
        "", headless=headless, use_uc=use_uc, page_load_timeout=page_load_timeout, script_timeout=script_timeout
    )
    try:
        executor.goto(url)
        html = executor.get_visible_html()
    finally:
        executor.quit()
    return html


def get_page_text(url, headless=False, use_uc=False, page_load_timeout=None, script_timeout=None):
    browser_print(f"Checking what {url} looks like...")
    executor = Executor(
        "", headless=headless, use_uc=use_uc, page_load_timeout=page_load_timeout, script_timeout=script_timeout
    )
    try:
        executor.goto(url)
        text = executor.driver.execute_script("return document.body.innerText;")
    finally:
        executor.quit()
    return text



import re

from rich.console import Console
from rich.panel import Panel
from rich.prompt import Prompt, Confirm
from rich.text import Text
from rich.progress import Progress as RichProgress

from parsagon.settings import GUI_ENABLED

if GUI_ENABLED:
    from parsagon.gui import GUIController

console = Console()


# === Printing ===
def normal_print(text):
    if not GUI_ENABLED:
        console.print(text)
    else:
        GUIController.shared().print(text)


def ask_reply():
    if not GUI_ENABLED:
        return ask("Reply or type Q to end")
    else:
        return ask(None)


def assistant_print(text):
    if not GUI_ENABLED:
        console.print(Panel(text, style="blue"))
    else:
        GUIController.shared().print(text, "black")


def browser_print(text):
    if not GUI_ENABLED:
        console.print(text, style="green")
    else:
        GUIController.shared().print(text, "green")


def error_print(text):
    if not GUI_ENABLED:
        console.print(text, style="red")
    else:
        GUIController.shared().print(text, "red")


# === Input, prompting ===
def ask(prompt, choices=None, show_choices=False):
    if not GUI_ENABLED:
        assert ask is not None
        return Prompt.ask(prompt, choices=choices, show_choices=show_choices)
    else:
        # TODO: Choices
        return GUIController.shared().input(prompt if prompt is not None else "")


original_input = input


def input(prompt):
    if not GUI_ENABLED:
        return original_input(prompt)
    else:
        return GUIController.shared().input(prompt)


def confirm(prompt):
    if not GUI_ENABLED:
        return Confirm.ask(prompt)
    else:
        return GUIController.shared().input(prompt + " (y/n):") == "y"


# === Status, loading ===
def assistant_spinner(text=""):
    if not GUI_ENABLED:
        return console.status(Text(text, style="blue"))
    else:
        return GUIController.shared().spinner()  # text argument


def status(text):
    if not GUI_ENABLED:
        return console.status(text)
    else:
        return GUIController.shared().spinner()  # text argument


class Progress:
    def __init__(self):
        if not GUI_ENABLED:
            self.rich_progress = RichProgress()

    def track(self, iterable, description="Working...", task_id=None):
        if not GUI_ENABLED:
            return self.rich_progress.track(iterable, description=description, task_id=task_id)
        else:
            description = re.sub(r"^\[[a-zA-Z]+]", "", description)

            def gui_generator():
                gui_controller = GUIController.shared()
                gui_controller.show_progress(True, description)
                gui_controller.progress_total = len(iterable)
                for i, item in enumerate(iterable):
                    yield item
                    GUIController.shared().set_progress(i)
                gui_controller.show_progress(False)

            return gui_generator()

    def add_task(self, description, total):
        if not GUI_ENABLED:
            return self.rich_progress.add_task(description, total=total)
        else:
            gui_controller = GUIController.shared()
            gui_controller.show_progress(True, description)
            gui_controller.progress_total = total

    def update(self, task_id, description):
        if not GUI_ENABLED:
            self.rich_progress.update(task_id, description=description)
        else:
            gui_controller = GUIController.shared()
            gui_controller.show_progress(False)
            gui_controller.print(description)

    def __enter__(self):
        if not GUI_ENABLED:
            return self.rich_progress.__enter__()
        else:
            return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Cleanup code, like closing files or releasing resources
        if not GUI_ENABLED:
            # Assuming RichProgress needs some kind of finalization
            self.rich_progress.__exit__(exc_type, exc_value, traceback)
        else:
            gui_controller = GUIController.shared()
            gui_controller.show_progress(False)
